# -*- coding: utf-8 -*-

"""Export Biomappings as SSSOM."""

import pathlib

import bioregistry
import click
import yaml

from biomappings import load_mappings, load_predictions
from biomappings.utils import DATA, MiriamValidator

DIRECTORY = pathlib.Path(DATA).joinpath("sssom")
DIRECTORY.mkdir(exist_ok=True, parents=True)
PATH = DIRECTORY.joinpath("biomappings.sssom.tsv")
META_PATH = DIRECTORY.joinpath("biomappings.sssom.yml")
META = {
    "license": "https://creativecommons.org/publicdomain/zero/1.0/",
    "mapping_provider": "https://github.com/biomappings/biomappings",
    "mapping_set_group": "biomappings",
    "mapping_set_id": "biomappings",
    "mapping_set_title": "Biomappings",
}

validator = MiriamValidator()


def get_sssom_df():
    """Get an SSSOM dataframe."""
    import pandas as pd

    rows = []
    prefixes = set()
    columns = [
        "subject_id",
        "predicate_id",
        "object_id",
        "subject_label",
        "object_label",
        "match_type",
        "creator_id",
        "confidence",
        "mapping_tool",
    ]
    for mapping in load_mappings():
        prefixes.add(mapping["source prefix"])
        prefixes.add(mapping["target prefix"])
        rows.append(
            (
                validator.get_curie(mapping["source prefix"], mapping["source identifier"]),
                f'{mapping["relation"]}',
                validator.get_curie(mapping["target prefix"], mapping["target identifier"]),
                mapping["source name"],
                mapping["target name"],
                "HumanCurated",  # match type
                mapping["source"],  # curator CURIE
                None,  # no confidence necessary
                None,  # mapping tool: none necessary for manually curated
            )
        )
    for mapping in load_predictions():
        prefixes.add(mapping["source prefix"])
        prefixes.add(mapping["target prefix"])
        rows.append(
            (
                validator.get_curie(mapping["source prefix"], mapping["source identifier"]),
                f'{mapping["relation"]}',
                validator.get_curie(mapping["target prefix"], mapping["target identifier"]),
                mapping["source name"],
                mapping["target name"],
                "LexicalEquivalenceMatch",  # match type
                None,  # no curator CURIE
                mapping["confidence"],
                mapping["source"],  # mapping tool: source script
            )
        )
    df = pd.DataFrame(rows, columns=columns)
    return prefixes, df


def get_msdf():
    """Get an SSSOM mapping set dataframe object."""
    # FIXME there are bugs in the linkml and SSSOM code that make this not work
    from sssom.datamodel_util import MappingSetDataFrame
    from sssom.parsers import from_dataframe

    _, df = get_sssom_df()
    msdf: MappingSetDataFrame = from_dataframe(
        df, curie_map=dict(bioregistry.get_format_urls()), meta=META
    )
    return msdf


@click.command()
@click.option("--path", default=PATH)
def sssom(path):
    """Export SSSOM."""
    prefixes, df = get_sssom_df()
    df.to_csv(path, sep="\t", index=False)

    # Get a CURIE map containing only the relevant prefixes
    curie_map = {
        prefix: formatter
        for prefix, formatter in bioregistry.get_format_urls().items()
        if prefix in prefixes
    }
    with open(META_PATH, "w") as file:
        yaml.safe_dump({"curie_map": curie_map, **META}, file)

    # TODO incorporate validation from sssom-py


if __name__ == "__main__":
    sssom()



# -*- coding: utf-8 -*-

"""Utilities."""

import os
import re
from subprocess import CalledProcessError, check_output  # noqa: S404
from typing import Any, Mapping, Optional, Tuple

import bioregistry
from bioregistry.external import get_miriam

HERE = os.path.dirname(os.path.abspath(__file__))
RESOURCE_PATH = os.path.abspath(os.path.join(HERE, "resources"))
DOCS = os.path.abspath(os.path.join(HERE, os.pardir, os.pardir, "docs"))
IMG = os.path.join(DOCS, "img")
DATA = os.path.join(DOCS, "_data")


def get_git_hash() -> Optional[str]:
    """Get the git hash.

    :return:
        The git hash, equals 'UNHASHED' if encountered CalledProcessError, signifying that the
        code is not installed in development mode.
    """
    rv = _git("rev-parse", "HEAD")
    if rv:
        return rv[:6]


def commit(message: str) -> Optional[str]:
    """Make a commit with the following message."""
    return _git("commit", "-m", message, "-a")


def push() -> Optional[str]:
    """Push the git repo."""
    return _git("push")


def not_main() -> bool:
    """Return if on the master branch."""
    return "master" != _git("rev-parse", "--abbrev-ref", "HEAD")


def _git(*args: str) -> Optional[str]:
    with open(os.devnull, "w") as devnull:
        try:
            ret = check_output(  # noqa: S603,S607
                ["git", *args],
                cwd=os.path.dirname(__file__),
                stderr=devnull,
            )
        except CalledProcessError:
            return
        else:
            return ret.strip().decode("utf-8")


def get_script_url(fname: str) -> str:
    """Get the source path for this script.

    :param fname: Pass ``__file__`` as the argument to this function.
    :return: The script's URL to GitHub
    """
    commit_hash = get_git_hash()
    script_name = os.path.basename(fname)
    return f"https://github.com/biomappings/biomappings/blob/{commit_hash}/scripts/{script_name}"


def get_canonical_tuple(mapping: Mapping[str, Any]) -> Tuple[str, str, str, str]:
    """Get the canonical tuple from a mapping entry."""
    source = mapping["source prefix"], mapping["source identifier"]
    target = mapping["target prefix"], mapping["target identifier"]
    if source > target:
        source, target = target, source
    return (*source, *target)


class InvalidPrefix(ValueError):
    """Raised for an invalid prefix."""


class InvalidIdentifier(ValueError):
    """Raised for an invalid identifier."""


class MiriamValidator:
    """Validate prefix/identifier pairs based on the MIRIAM database."""

    def __init__(self, force_download: bool = False):  # noqa: D107
        self.entries = self._load_identifiers_entries(force_download=force_download)

    @staticmethod
    def _load_identifiers_entries(force_download: bool = False):
        return {
            prefix: {
                "pattern": re.compile(entry["pattern"]),
                "namespace_embedded": entry["namespaceEmbeddedInLui"],
            }
            for prefix, entry in get_miriam(force_download=force_download).items()
        }

    def namespace_embedded(self, prefix: str) -> bool:
        """Return True if the namespace is embedded for the given prefix."""
        if prefix in self.entries:
            return self.entries[prefix]["namespace_embedded"]
        return bioregistry.namespace_in_lui(prefix)

    def check_valid_prefix_id(self, prefix, identifier):
        """Check the prefix/identifier pair is valid."""
        if prefix in self.entries:
            entry = self.entries[prefix]
            if not re.match(entry["pattern"], identifier):
                raise InvalidIdentifier(prefix, identifier)
        elif bioregistry.get(prefix) is None:
            raise InvalidPrefix(prefix)
        elif bioregistry.get_pattern(prefix) is None:
            if bioregistry.validate(prefix, identifier):
                raise InvalidIdentifier(prefix, identifier)

    def get_curie(self, prefix: str, identifier: str) -> str:
        """Return CURIE for a given prefix and identifier."""
        if self.namespace_embedded(prefix):
            return identifier
        else:
            return f"{prefix}:{identifier}"

    @staticmethod
    def get_url(prefix: str, identifier: str) -> str:
        """Return URL for a given prefix and identifier."""
        return bioregistry.get_link(prefix, identifier, use_bioregistry_io=False)



# -*- coding: utf-8 -*-

"""The biomappings CLI."""

import click
from more_click import verbose_option

from .export_sssom import sssom
from .graph import charts
from .summary import export
from .upload_ndex import ndex


@click.group()
def main():
    """Run the biomappings CLI."""


@main.command()
@verbose_option
def web():
    """Run the biomappings web curation interface."""
    from .wsgi import app
    from .utils import get_git_hash

    if get_git_hash() is None:
        click.secho(
            "You are not running biomappings from a development installation.\n"
            "Please run the following to install in development mode:\n"
            "  $ git clone https://github.com/biomappings/biomappings.git\n"
            "  $ cd biomappings\n"
            "  $ pip install -e .",
            fg="red",
        )
    else:
        app.run()


@main.command()
@click.pass_context
def update(ctx: click.Context):
    """Run all update functions."""
    click.secho("Building general exports", fg="green")
    ctx.invoke(export)
    click.secho("Building SSSOM export", fg="green")
    ctx.invoke(sssom)
    click.secho("Generating charts", fg="green")
    ctx.invoke(charts)
    click.secho("Uploading to NDEx", fg="green")
    ctx.invoke(ndex)


main.add_command(export)
main.add_command(ndex)
main.add_command(charts)

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

"""Load Biomappings as a graph."""

import os
from collections import Counter
from operator import itemgetter
from typing import Iterable, List, Mapping, Optional, Sequence

import click
import networkx as nx
import yaml

from biomappings.resources import load_false_mappings, load_mappings
from biomappings.utils import DATA, IMG, MiriamValidator


def get_true_graph(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> nx.Graph:
    """Get a graph of the true mappings."""
    return _graph_from_mappings(load_mappings(), include=include, exclude=exclude)


def get_false_graph(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> nx.Graph:
    """Get a graph of the false mappings."""
    return _graph_from_mappings(load_false_mappings(), include=include, exclude=exclude)


def get_predictions_graph(
    include: Optional[Sequence[str]] = None, exclude: Optional[Sequence[str]] = None
) -> nx.Graph:
    """Get a graph of the predicted mappings."""
    return _graph_from_mappings(load_false_mappings(), include=include, exclude=exclude)


def _graph_from_mappings(
    mappings: Iterable[Mapping[str, str]],
    include: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
) -> nx.Graph:
    v = MiriamValidator()
    graph = nx.Graph()

    if include is not None:
        include = set(include)
        print("only including", *include)
    if exclude is not None:
        exclude = set(exclude)
        print("excluding", *exclude)

    for mapping in mappings:
        relation = mapping["relation"]
        if exclude and (relation in exclude):
            continue
        if include and (relation not in include):
            continue

        source_curie = v.get_curie(mapping["source prefix"], mapping["source identifier"])
        graph.add_node(
            source_curie,
            prefix=mapping["source prefix"],
            identifier=mapping["source identifier"],
            name=mapping["source name"],
        )
        target_curie = v.get_curie(mapping["target prefix"], mapping["target identifier"])
        graph.add_node(
            target_curie,
            prefix=mapping["target prefix"],
            identifier=mapping["target identifier"],
            name=mapping["target name"],
        )
        graph.add_edge(
            source_curie,
            target_curie,
            relation=relation,
            provenance=mapping["source"],
            type=mapping["type"],
        )
    return graph


@click.command()
def charts():
    """Make charts."""
    import matplotlib.pyplot as plt
    import seaborn as sns

    miriam_validator = MiriamValidator()
    true_mappings = load_mappings()
    true_graph = _graph_from_mappings(true_mappings, include=["skos:exactMatch"])

    component_node_sizes, component_edge_sizes, component_densities, component_number_prefixes = (
        [],
        [],
        [],
        [],
    )
    prefix_list = []
    components_with_duplicate_prefixes = []
    incomplete_components = []
    n_duplicates = []
    for component in nx.connected_components(true_graph):
        component = true_graph.subgraph(component)
        node_size = component.number_of_nodes()
        edge_size = component.number_of_edges()

        nodes_data = {
            curie: {
                "link": miriam_validator.get_url(data["prefix"], data["identifier"]),
                **data,
            }
            for curie, data in sorted(component.nodes(data=True), key=itemgetter(0))
        }

        component_node_sizes.append(node_size)
        component_edge_sizes.append(edge_size)
        if node_size > 2:
            component_densities.append(nx.density(component))
        if node_size > 2 and edge_size < (node_size * (node_size - 1) / 2):
            incomplete_components_edges = []
            for u, v in sorted(nx.complement(component.copy()).edges()):
                if u > v:
                    u, v = v, u
                incomplete_components_edges.append(
                    {
                        "source": {"curie": u, **nodes_data[u]},
                        "target": {"curie": v, **nodes_data[v]},
                    }
                )
            incomplete_components_edges = sorted(
                incomplete_components_edges, key=lambda d: d["source"]["curie"]
            )
            incomplete_components.append(
                {
                    "nodes": nodes_data,
                    "edges": incomplete_components_edges,
                }
            )

        prefixes = [true_graph.nodes[node]["prefix"] for node in component]
        prefix_list.extend(prefixes)
        unique_prefixes = len(set(prefixes))
        component_number_prefixes.append(unique_prefixes)
        _n_duplicates = len(prefixes) - unique_prefixes
        n_duplicates.append(_n_duplicates)
        if _n_duplicates:
            components_with_duplicate_prefixes.append(nodes_data)

    with open(os.path.join(DATA, "incomplete_components.yml"), "w") as file:
        yaml.safe_dump(incomplete_components, file)
    with open(os.path.join(DATA, "components_with_duplicate_prefixes.yml"), "w") as file:
        yaml.safe_dump(components_with_duplicate_prefixes, file)

    fig, axes = plt.subplots(2, 3, figsize=(10.5, 6.5))

    _countplot_list(component_node_sizes, ax=axes[0][0])
    axes[0][0].set_yscale("log")
    axes[0][0].set_title("Size (Nodes)")

    _countplot_list(component_edge_sizes, ax=axes[0][1])
    axes[0][1].set_yscale("log")
    axes[0][1].set_title("Size (Edges)")
    axes[0][1].set_ylabel("")

    sns.kdeplot(component_densities, ax=axes[0][2])
    axes[0][2].set_xlim([0.0, 1.0])
    # axes[0][2].set_yscale('log')
    axes[0][2].set_title("Density ($|V| > 2$)")
    axes[0][2].set_ylabel("")

    _countplot_list(component_number_prefixes, ax=axes[1][0])
    axes[1][0].set_title("Number Prefixes")
    axes[1][0].set_yscale("log")
    # has duplicate prefix in component

    _countplot_list(n_duplicates, ax=axes[1][1])
    axes[1][1].set_yscale("log")
    axes[0][2].set_ylabel("")
    axes[1][1].set_title("Number Duplicate Prefixes")

    axes[1][2].axis("off")

    path = os.path.join(IMG, "components.png")
    print("saving to", path)
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.close(fig)

    fig, axes = plt.subplots(1, 2, figsize=(8, 3.5))
    sns.countplot(
        y=prefix_list, ax=axes[0], order=[k for k, _ in Counter(prefix_list).most_common()]
    )
    axes[0].set_xscale("log")
    axes[0].set_title("Prefix Frequency")

    relations = [m["relation"] for m in true_mappings]
    sns.countplot(y=relations, ax=axes[1], order=[k for k, _ in Counter(relations).most_common()])
    axes[1].set_xscale("log")
    axes[1].set_title("Relation Frequency")

    path = os.path.join(IMG, "summary.png")
    print("saving to", path)
    plt.tight_layout()
    plt.savefig(path, dpi=300)
    plt.close(fig)


def _countplot_list(data: List[int], ax):
    import pandas as pd
    import seaborn as sns

    counter = Counter(data)
    for size in range(min(counter), max(counter)):
        if size not in counter:
            counter[size] = 0
    df = pd.DataFrame(counter.items(), columns=["size", "count"]).sort_values("size").reset_index()
    sns.barplot(data=df, x="size", y="count", ax=ax)


if __name__ == "__main__":
    charts()



# -*- coding: utf-8 -*-

"""Web curation interface for :mod:`biomappings`."""

import getpass
import os
from collections import defaultdict
from typing import Any, Iterable, Mapping, Optional, Tuple

import flask
import flask_bootstrap
from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField

from biomappings.resources import (
    append_false_mappings,
    append_true_mappings,
    append_unsure_mappings,
    load_curators,
    load_predictions,
    write_predictions,
)
from biomappings.utils import MiriamValidator, commit, not_main, push

app = flask.Flask(__name__)
app.config["WTF_CSRF_ENABLED"] = False
app.config["SECRET_KEY"] = os.urandom(8)
app.config["SHOW_RELATIONS"] = True
flask_bootstrap.Bootstrap(app)

# A mapping from your computer's user, returned by getuser.getpass()
KNOWN_USERS = {record["user"]: record["orcid"] for record in load_curators()}


def _manual_source():
    known_user = KNOWN_USERS.get(getpass.getuser())
    if known_user:
        return f"orcid:{known_user}"
    return "web"


miriam_validator = MiriamValidator()


class Controller:
    """A module for interacting with the predictions and mappings."""

    def __init__(self):  # noqa: D107
        self._predictions = load_predictions()
        self._marked = {}
        self.total_curated = 0
        self._added_mappings = []

    def predictions(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        query: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> Iterable[Tuple[int, Mapping[str, Any]]]:
        """Iterate over predictions.

        :param offset: If given, offset the iteration by this number
        :param limit: If given, only iterate this number of predictions.
        :param query: If given, show only equivalences that have it appearing as a substring in one of the fields
        :param prefix: If given, show only equivalences that have it appearing as a substring in one of the prefixes
        :yields: Pairs of positions and prediction dictionaries
        """
        it = self._help_it_predictions(query=query, prefix=prefix)
        if offset is not None:
            try:
                for _ in range(offset):
                    next(it)
            except StopIteration:
                # if next() fails, then there are no remaining entries.
                # do not pass go, do not collect 200 euro $
                return
        if limit is None:
            yield from it
        else:
            for line_prediction, _ in zip(it, range(limit)):
                yield line_prediction

    def count_predictions(
        self,
        query: Optional[str] = None,
        prefix: Optional[str] = None,
    ) -> int:
        """Count the number of predictions to check for the given filters."""
        return sum(1 for _ in self._help_it_predictions(query=query, prefix=prefix))

    def _help_it_predictions(self, query: Optional[str] = None, prefix: Optional[str] = None):
        it = enumerate(self._predictions)
        if query is not None:
            query = query.casefold()
            it = (
                (line, prediction)
                for line, prediction in it
                if any(
                    query in prediction[x].casefold()
                    for x in (
                        "source identifier",
                        "source name",
                        "target identifier",
                        "target name",
                    )
                )
            )
        if prefix is not None:
            prefix = prefix.casefold()
            it = (
                (line, prediction)
                for line, prediction in it
                if any(
                    prefix in prediction[x].casefold() for x in ("source prefix", "target prefix")
                )
            )

        it = ((line, prediction) for line, prediction in it if line not in self._marked)
        return it

    @staticmethod
    def get_curie(prefix: str, identifier: str) -> str:
        """Return CURIE for a given prefix and identifier."""
        return miriam_validator.get_curie(prefix, identifier)

    @classmethod
    def get_url(cls, prefix: str, identifier: str) -> str:
        """Return URL for a given prefix and identifier."""
        return miriam_validator.get_url(prefix, identifier)

    @property
    def total_predictions(self) -> int:
        """Return the total number of yet unmarked predictions."""
        return len(self._predictions) - len(self._marked)

    def mark(self, line: int, value: str) -> None:
        """Mark the given equivalency as correct.

        :param line: Position of the prediction
        :param value: Value to mark the prediction with
        :raises ValueError: if an invalid value is used
        """
        if line not in self._marked:
            self.total_curated += 1
        if value not in {"correct", "incorrect", "unsure"}:
            raise ValueError
        self._marked[line] = value

    def add_mapping(
        self,
        source_prefix: str,
        source_id: str,
        source_name: str,
        target_prefix: str,
        target_id: str,
        target_name: str,
    ) -> None:
        """Add manually curated new mappings."""
        try:
            miriam_validator.check_valid_prefix_id(source_prefix, source_id)
        except ValueError as e:
            flask.flash(
                f"Problem with source CURIE {source_prefix}:{source_id}: {e.__class__.__name__}",
                category="warning",
            )
            return

        try:
            miriam_validator.check_valid_prefix_id(target_prefix, target_id)
        except ValueError as e:
            flask.flash(
                f"Problem with target CURIE {target_prefix}:{target_id}: {e.__class__.__name__}",
                category="warning",
            )
            return

        self._added_mappings.append(
            {
                "source prefix": source_prefix,
                "source identifier": source_id,
                "source name": source_name,
                "relation": "skos:exactMatch",
                "target prefix": target_prefix,
                "target identifier": target_id,
                "target name": target_name,
                "source": _manual_source(),
                "type": "manual",
            }
        )
        self.total_curated += 1

    def persist(self):
        """Save the current markings to the source files."""
        entries = defaultdict(list)

        for line, value in sorted(self._marked.items(), reverse=True):
            prediction = self._predictions.pop(line)
            prediction["source"] = _manual_source()
            prediction["type"] = "manually_reviewed"
            entries[value].append(prediction)

        append_true_mappings(entries["correct"])
        append_false_mappings(entries["incorrect"])
        append_unsure_mappings(entries["unsure"])
        write_predictions(self._predictions)
        self._marked.clear()

        # Now add manually curated mappings
        append_true_mappings(self._added_mappings)
        self._added_mappings = []


controller = Controller()


class MappingForm(FlaskForm):
    """Form for entering new mappings."""

    source_prefix = StringField("Source Prefix", id="source_prefix")
    source_id = StringField("Source ID", id="source_id")
    source_name = StringField("Source Name", id="source_name")
    target_prefix = StringField("Target Prefix", id="target_prefix")
    target_id = StringField("Target ID", id="target_id")
    target_name = StringField("Target Name", id="target_name")
    submit = SubmitField("Add")


@app.route("/")
def home():
    """Serve the home page."""
    form = MappingForm()
    limit = flask.request.args.get("limit", type=int, default=10)
    offset = flask.request.args.get("offset", type=int, default=0)
    query = flask.request.args.get("query")
    prefix = flask.request.args.get("prefix")
    show_relations = app.config["SHOW_RELATIONS"]
    return flask.render_template(
        "home.html",
        controller=controller,
        form=form,
        limit=limit,
        offset=offset,
        query=query,
        prefix=prefix,
        show_relations=show_relations,
    )


@app.route("/add_mapping", methods=["POST"])
def add_mapping():
    """Add a new mapping manually."""
    form = MappingForm()
    if form.is_submitted():
        controller.add_mapping(
            form.data["source_prefix"],
            form.data["source_id"],
            form.data["source_name"],
            form.data["target_prefix"],
            form.data["target_id"],
            form.data["target_name"],
        )
        controller.persist()
    else:
        flask.flash("missing form data", category="warning")
    return _go_home()


@app.route("/commit")
def run_commit():
    """Make a commit then redirect to the the home page."""
    commit(
        f'Curated {controller.total_curated} mapping{"s" if controller.total_curated > 1 else ""}'
        f" ({getpass.getuser()})",
    )
    if not_main():
        push()
    controller.total_curated = 0
    return _go_home()


CORRECT = {"yup", "true", "t", "correct", "right", "close enough", "disco"}
INCORRECT = {"no", "nope", "false", "f", "nada", "nein", "incorrect", "negative", "negatory"}
UNSURE = {"unsure", "maybe", "idk", "idgaf", "idgaff"}


def _normalize_mark(value: str) -> str:
    value = value.lower()
    if value in CORRECT:
        return "correct"
    elif value in INCORRECT:
        return "incorrect"
    elif value in UNSURE:
        return "unsure"
    else:
        raise ValueError


@app.route("/mark/<int:line>/<value>")
def mark(line: int, value: str):
    """Mark the given line as correct or not."""
    controller.mark(line, _normalize_mark(value))
    controller.persist()
    return _go_home()


def _go_home():
    return flask.redirect(
        flask.url_for(
            "home",
            limit=flask.request.args.get("limit", type=int),
            offset=flask.request.args.get("offset", type=int),
            query=flask.request.args.get("query"),
            prefix=flask.request.args.get("prefix"),
            show_relations=app.config["SHOW_RELATIONS"],
        )
    )


if __name__ == "__main__":
    app.run()
